/*
	Copyright (c) 2004 Jonatan Kronqvist

	Permission is hereby granted, free of charge, to any person obtaining a
	copy of this software and associated documentation files (the "Software"),
	to deal in the Software without restriction, including without limitation
	the rights to use, copy, modify, merge, publish, distribute, sublicense,
	and/or sell copies of the Software, and to permit persons to whom the
	Software is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
	OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
	FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
	IN THE SOFTWARE.
*/


#include "bluetooth_control.h"


bluetooth_control::bluetooth_control(connection *conn, control *play_control)
{
	m_conn = conn;
	m_play_control = play_control;
	m_activated = false;
	m_dirlist_thread = NULL;
}


bluetooth_control::~bluetooth_control()
{
	SAFE_DELETE(m_conn);
}


bool bluetooth_control::set_active(bool activate)
{
	if(!activate) {
		if(m_dirlist_thread != NULL) {
			WaitForSingleObject(m_dirlist_thread, 0xFFFFFFFF);
			CloseHandle(m_dirlist_thread);
			m_dirlist_thread = NULL;
		}
	}
	return true;
}

void bluetooth_control::set_musicdir(string8 musicdir)
{
	m_musicdir = musicdir;
}
void bluetooth_control::set_random(bool random)
{
	m_random = random;
}



void bluetooth_control::send_command(unsigned long cmd)
{
	process_command(cmd);
}



unsigned long WINAPI bluetooth_control::thread_send_entire_dirlist(LPVOID param)
{
	bluetooth_control *ctrl = reinterpret_cast<bluetooth_control*>(param);
	ctrl->send_entire_list();
	return 0;
}



string8 bluetooth_control::read_filename()
{
	string8 fname;
	unsigned short flen = 0;
	if(m_conn->read(&flen, 2) != 2)
		console::error("Read error!");

	flen = SWAP_SHORT(flen);
	char *filename = new char[flen];
	if(m_conn->read((void*)filename, flen) != flen)
		console::error("Read error!");

	fname.add_string(filename, flen);
	delete filename;
	return fname;
}


/**
 * The Bemused protocol can be found at:
 * http://www.compsoc.man.ac.uk/~ashley/bemused/BemusedProtocol.html
 */
void bluetooth_control::process_command(unsigned long _cmd)
{
	if(!core_api::assert_main_thread()) {
		console::error("process_command() called from outside of main thread.");
		return;
	}

	// FIXME: this could be done prettier by not converting back to string

	char cmd[5];
	memcpy(cmd, &_cmd, 4);

	string8 str ("Command recieved: ");
	str.add_string(cmd,4);
	g_debug(str);

	if(!strncmp(cmd, "CHCK", 4)) {
		m_conn->send("Y", 1);
	}
	else if(!strncmp(cmd, "DINF", 4)) {
		send_detailed_info();
	}
	else if(!strncmp(cmd, "DLST", 4)) {
		send_partial_list(read_filename());
	}
	else if(!strncmp(cmd, "DOWN", 4)) {
		// downloading of files not supported
	}
	else if(!strncmp(cmd, "FADE", 4)) {
		m_play_control->fade_stop();
	}
	else if(!strncmp(cmd, "FFWD", 4)) {
		m_play_control->ffwd(5);
	}
	else if(!strncmp(cmd, "FINF", 4)) {
		send_fileinfo(read_filename());
	}
	else if(!strncmp(cmd, "GVOL", 4)) {
		send_volume();
	}
	else if(!strncmp(cmd, "INFO", 4)) {
		send_songinfo();
	}
	else if(!strncmp(cmd, "INF2", 4)) {
		send_songinfo2();
	}
	else if(!strncmp(cmd, "LADD", 4)) {
		m_play_control->add_to_playlist(read_filename());
	}
	else if(!strncmp(cmd, "LIST", 4)) {
		if(m_dirlist_thread != NULL) {
			WaitForSingleObject(m_dirlist_thread, INFINITE);	// block until other send finished if already started
			CloseHandle(m_dirlist_thread);
		}
		m_dirlist_thread = CreateThread(NULL, 0, bluetooth_control::thread_send_entire_dirlist, reinterpret_cast<void*>(this), 0, NULL);
		SetThreadPriority(m_dirlist_thread, THREAD_PRIORITY_HIGHEST);
	}
	else if(!strncmp(cmd, "NEXT", 4)) {
		m_play_control->next(m_random);
	}
	else if(!strncmp(cmd, "PAUS", 4)) {
		m_play_control->pause();
	}
	else if(!strncmp(cmd, "PLAY", 4)) {
		m_play_control->play(read_filename());
	}
	else if(!strncmp(cmd, "PLEN", 4)) {
		send_playlist_length();
	}
	else if(!strncmp(cmd, "PLST", 4)) {
		send_playlist();
	}
	else if(!strncmp(cmd, "PREV", 4)) {
		m_play_control->prev();
	}
	else if(!strncmp(cmd, "REPT", 4)) {
		char repeat=0;
		m_conn->read(&repeat, 1);
		m_play_control->set_repeat( repeat == 1 );
	}
	else if(!strncmp(cmd, "RMAL", 4)) {
		m_play_control->clear_playlist();
	}
	else if(!strncmp(cmd, "RWND", 4)) {
		m_play_control->rwnd(5);
	}
	else if(!strncmp(cmd, "SHFL", 4)) {
		char shuffle=0;
		m_conn->read(&shuffle, 1);
		m_play_control->set_shuffle( shuffle == 1 );
	}
	else if(!strncmp(cmd, "SEEK", 4)) {
		long pos;
		m_conn->read(&pos, 4);
		m_play_control->seek(SWAP_LONG(pos));
	}
	else if(!strncmp(cmd, "SHUT", 4)) {
		// shut down PC.. not supported
	}
	else if(!strncmp(cmd, "STEN", 4)) {
		m_play_control->set_stop_at_eof_track(true);
	}
	else if(!strncmp(cmd, "SLCT", 4)) {
		short idx;
		m_conn->read(&idx, 2);
		m_play_control->select_at(SWAP_SHORT(idx));
	}
	else if(!strncmp(cmd, "STOP", 4)) {
		m_play_control->stop();
	}
	else if(!strncmp(cmd, "STRT", 4)) {
		m_play_control->play_current();
	}
	else if(!strncmp(cmd, "VOLM", 4)) {
		char vol;
		m_conn->read(&vol, 1);
		m_play_control->set_volume(vol);
	}
	else if(!strncmp(cmd, "VERS", 4)) {
		send_version();
	}
}


void bluetooth_control::send_detailed_info()
{
	g_debug("Sending detailed info");

	/* format of response:
		length	field name
		7		"DINFACK"
		4		bitrate
		4		samplerate
		4		number of channels
	*/
	long bitrate = SWAP_LONG(m_play_control->get_bitrate());
	long samplerate = SWAP_LONG(m_play_control->get_samplerate());
	long channels = SWAP_LONG(m_play_control->get_channels());

	buffer buf(19);
	buf.append("DINFACK");
	buf.append(bitrate);
	buf.append(samplerate);
	buf.append(channels);
	m_conn->send(buf);
}


void bluetooth_control::send_dirlist(DirTreeNode *root)
{
	if(root == NULL) return;

	buffer buf(root->name.length()+2);
	buf.append(root->type);
	buf.append(root->name);
	buf.append('\0');
	m_conn->send(buf);

	string8 tmpstr;
	tmpstr.add_int(root->type >> 4);
	tmpstr.add_char(' ');
	tmpstr.add_int(root->type & 0xF);
	tmpstr.add_char(' ');
	tmpstr.add_string(root->name);
	g_debug(tmpstr);

	send_dirlist(root->first_child);
	for(unsigned int i=0; i<root->siblings.size(); i++)
		send_dirlist(root->siblings[i]);
}


DirTreeNode *bluetooth_control::build_dirtree(string8 root, string8 prefix, unsigned char type, bool recursive)
{
	uFindFile	*find;
	string8		filename;
	// cur is pointing to a dir with name in root
	DirTreeNode	*cur = new DirTreeNode(root, type);

	string8 path, glob;
	path.add_string(prefix);
	path.add_string(root);
	path.add_string("\\");

	glob.set_string(path);
	glob.add_string("*.*");
	find = uFindFirstFile(glob);

	// TODO: reorganise this loop..
	do {
		filename.set_string(find->GetFileName());
		if(find->IsDirectory()) {
			if(strncmp(find->GetFileName(), ".", 1)) {	// don't do anything for directories starting with "." e.g. "." and ".."
				if(recursive) {
					if(cur->first_child == NULL)
						cur->first_child = build_dirtree(filename, path, DIRLST_CHILD+DIRLST_DIR, true);
					else
						cur->siblings.push_back(build_dirtree(filename, path, DIRLST_SIBLING+DIRLST_DIR, true));
				} else {
					if(cur->first_child == NULL)
						cur->first_child = new DirTreeNode(filename, DIRLST_CHILD + DIRLST_UXDIR);
					else
						cur->siblings.push_back(new DirTreeNode(filename, DIRLST_SIBLING + DIRLST_UXDIR));
				}
			}
		} else {
			// normal file
			if(cur->first_child == NULL)
				cur->first_child = new DirTreeNode(filename, DIRLST_CHILD + DIRLST_FILE);
			else
				cur->siblings.push_back(new DirTreeNode(filename, DIRLST_SIBLING + DIRLST_FILE));
		}
	} while(find->FindNext());

	if(cur->siblings.empty()) {		// last is a child, mark as only child
		if(cur->first_child != NULL)
			cur->first_child->type = DIRLST_OCHILD + (cur->first_child->type & 0x0F);
	} else {	// last is a sibling, mark as last sibling
		cur->siblings.back()->type = DIRLST_LSIBLING + (cur->siblings.back()->type & 0x0F);
	}

	delete find;

	return cur;
}


void bluetooth_control::delete_dirtree(DirTreeNode *root)
{
	if(root->first_child != NULL)
		delete_dirtree(root->first_child);

	for(unsigned int i=0; i<root->siblings.size(); i++) {
		delete_dirtree(root->siblings[i]);
	}
	root->siblings.clear();

	SAFE_DELETE(root);
}






void bluetooth_control::send_fileinfo(string8 fname)
{
	long size = 0;

	uFindFile *find = uFindFirstFile(m_play_control->get_file_path());
	size = SWAP_LONG((long)find->GetFileSize());
	delete find;

	buffer buf(11);
	buf.append("FINFACK");
	buf.append(size);
	m_conn->send(buf);
}


void bluetooth_control::send_volume()
{
	/* format of response:
		length	field name
		7		"GVOLACK"
		1		volume
	*/

	unsigned char vol = m_play_control->get_volume();

	string8 tmpstr ("Sending vol: ");
	tmpstr.add_int(vol);
	g_debug(tmpstr);

	buffer buf(8);
	buf.append("VOLACK");
	buf.append(vol);
	m_conn->send(buf);
}



void bluetooth_control::send_songinfo()
{
	g_debug("Sending songinfo");

	/* format of the response:
		length	field name
		7 		"INFOACK"
		1		play state
		4		song length in seconds
		4		current time played in seconds
		1		shuffle state
		1		repeat state
	*/
	char playstate = 0;
	if(m_play_control->is_playing()) playstate = playstate | 0x1;
	if(m_play_control->is_paused()) playstate = playstate | 0x2;

	long playlen = SWAP_LONG(m_play_control->get_playlength());
	long playtime = SWAP_LONG(m_play_control->get_playback_time());

	char shuffle = m_play_control->get_shuffle() ? 1 : 0;
	char repeat = m_play_control->get_repeat() ? 1 : 0;

	string8 name;
	m_play_control->get_title(name);
	name.add_char('\0');

	buffer buf(18+name.length());
	buf.append("INFOACK");
	buf.append(playstate);
	buf.append(playlen);
	buf.append(playtime);
	buf.append(shuffle);
	buf.append(repeat);
	buf.append(name);
	m_conn->send(buf);
}


void bluetooth_control::send_songinfo2()
{
	g_debug("Sending songinfo2");
	send_songinfo();
	m_conn->send("\0", 1);
}



void bluetooth_control::send_entire_list()
{
	string8 tmprootdir (m_musicdir);
	DirTreeNode *root = NULL;

	int nextsep = tmprootdir.find_first(';');
	if(nextsep != -1) {
		// We have multiple roots
		root = new DirTreeNode("", DIRLST_ROOT+DIRLST_DIR);
		int dirs = 0;
		while(nextsep != -1) {
			string8 rootdir = tmprootdir;
			nextsep = tmprootdir.find_first(';');		// find next separator
			if(nextsep != -1) {
				rootdir.truncate(nextsep);
				tmprootdir.remove_chars(0,nextsep+1);		// slow..
			}
			if(dirs == 0)
				root->first_child = build_dirtree(rootdir, "", DIRLST_CHILD+DIRLST_DIR, true);
			else
				root->siblings.push_back(build_dirtree(rootdir, "", DIRLST_SIBLING+DIRLST_DIR, true));
			dirs++;
		}
		// OCHILD will never be needed, but set last sibling to LSIBLING here
		root->siblings.back()->type = DIRLST_LSIBLING + DIRLST_DIR;
	} else {
		// only one root
		root = build_dirtree(tmprootdir, "", DIRLST_ROOT+DIRLST_DIR, true);
	}

	m_conn->send("LISTACK", 7);
	send_dirlist(root);

	unsigned char terminator[2];
	terminator[0] = 0xFF;
	terminator[1] = 0x00;
	m_conn->send(terminator, 2);		// send terminating node

	delete_dirtree(root);
}


void bluetooth_control::send_partial_list(string8 dir)
{
	DirTreeNode *root = build_dirtree(dir, "", DIRLST_ROOT+DIRLST_DIR, false);

	m_conn->send("LISTACK", 7);
	send_dirlist(root);

	unsigned char terminator[2];
	terminator[0] = 0xFF;
	terminator[1] = 0x00;
	m_conn->send(terminator, 2);		// send terminating node

	delete_dirtree(root);
}





void bluetooth_control::send_playlist_length()
{
	short playlist_len = (short)m_play_control->get_playlist_len();
	if(playlist_len > 65535) {		// report overflow
		console::error("The playlist exceeds 65535 songs, it cannot be uploaded to the phone!");
	}

	playlist_len = SWAP_SHORT(playlist_len);
	m_conn->send(&playlist_len, 2);
}


void bluetooth_control::send_playlist()
{
	g_debug("Sending playlist");
	m_conn->send("PLSTACK", 7);

	if(m_play_control->get_playlist_len() > 65535) {		// report overflow
		console::error("The playlist exceeds 65535 songs, it cannot be uploaded to the phone!");
		m_conn->send("\0", 1);
		return;
	}

	unsigned short cur_idx = SWAP_SHORT((unsigned short)m_play_control->get_cur_idx());
	m_conn->send(&cur_idx, 2);

	std::vector<string8> *playlist = m_play_control->get_playlist();

	for(unsigned int i=0; i<playlist->size(); i++)
		m_conn->send((void*)(playlist->at(i).get_ptr()), playlist->at(i).length());
	m_conn->send("\0", 1);
	g_debug("Playlist sent");
}



void bluetooth_control::send_version()
{
	g_debug("Sending version info to phone.");

	unsigned char ver[2];
	ver[0] = 1;
	ver[1] = 73;
	m_conn->send(ver, 2);
}

