/*
	Copyright (c) 2004 Jonatan Kronqvist

	Permission is hereby granted, free of charge, to any person obtaining a
	copy of this software and associated documentation files (the "Software"),
	to deal in the Software without restriction, including without limitation
	the rights to use, copy, modify, merge, publish, distribute, sublicense,
	and/or sell copies of the Software, and to permit persons to whom the
	Software is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
	OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
	FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
	IN THE SOFTWARE.
*/


#include "foo_bluetooth_ctrl.h"
#include "buffer.h"



foo_bluetooth_ctrl::foo_bluetooth_ctrl()
{
	console::info("Initializing");
	m_conn = new bluetooth_connection();
	m_play_control = new foobar_control();
	m_bt_control = new bluetooth_control(m_conn, m_play_control);
	m_activated = false;
	m_io_thread = NULL;
}


foo_bluetooth_ctrl::~foo_bluetooth_ctrl()
{
	console::info("Terminating");
	set_active(false);	// deactivate
	SAFE_DELETE(m_bt_control);
	SAFE_DELETE(m_play_control);
}


void foo_bluetooth_ctrl::set_active(bool activate)
{
	m_activated = activate;
	if(activate) {
		// Activate == run the component
		// create a window for us to send messages to
		uWNDCLASS wc;
		memset(&wc,0,sizeof(wc));
		wc.style = 0;
		wc.lpfnWndProc = foo_bluetooth_ctrl::wndproc;
		wc.hInstance = core_api::get_my_instance();
		wc.lpszClassName = "foo_bluetooth_ctrl";
		uRegisterClass(&wc);

		m_hwnd = uCreateWindowEx(0,
								"foo_bluetooth_ctrl",
								"foo_bluetooth_ctrl",
								0/*WS_OVERLAPPED|WS_CAPTION*/,
								CW_USEDEFAULT,
								CW_USEDEFAULT,
								0,0,
								core_api::get_main_window(),
								0,
								core_api::get_my_instance(),
								0);
		if(!m_hwnd) console::error("couldn't create window");

		m_bt_control->set_active(true);
		update_config();

		m_conn->set_parameter(cfg_comport.get_val());
		// try to connect
		if(!m_conn->connect()) {
			console::beep();
			console::error("Could not connect!");
			console::popup();
			set_active(false);
			return;
		}

		m_io_thread = CreateThread(NULL, 0, foo_bluetooth_ctrl::read_command_thread, reinterpret_cast<void*>(this), 0, NULL);

		console::info("Activated");
	} else {
		// Deactivate
		uDestroyWindow(m_hwnd);
		if(m_io_thread != NULL) {
			TerminateThread(m_io_thread, 0);	// FIXME: this is not very nice
			CloseHandle(m_io_thread);
			m_io_thread = NULL;
		}
		m_bt_control->set_active(false);
		m_conn->disconnect();
		console::info("Deactivated");
	}
}

void foo_bluetooth_ctrl::update_config()
{
	m_bt_control->set_musicdir(cfg_musicdir.get_val());
	m_bt_control->set_random((cfg_random.get_value() != 0));
	foobar_control* tmp = reinterpret_cast<foobar_control*>(m_play_control);
	tmp->set_playlist_fmt(cfg_playlist_fmt.get_val());
}


bool foo_bluetooth_ctrl::is_active()
{
	return m_activated;
}


void foo_bluetooth_ctrl::read_command()
{
	// all commands are 4 bytes long, which means that we can directly map
	// them to an unsigned long for instance. This very nasty hack simplifies
	// the passing of commands immensely, since we won't have to deal with (de)allocating
	// the same string in different threads
	unsigned long cmd;
	while(m_activated) {
		if(m_conn->read(&cmd, 4) != 4) {
			continue;
//			console::error("Read error!");
		}

		if(cmd == 0x4F464E49)	// == INFO
			Sleep(300);			// FIXME: ugly hack to allow foobar to actually start playing the song before
								// we try to read info about it and send to the phone. If we don't sleep the
								// info sent will be wrong.

		// send a message with the command, this to be sure that the command is
		// executed from the main thread to avoid deadlocks and other threading issues.
		uSendMessage(m_hwnd, WM_BT_CMD, 0, cmd);
		Sleep(100);		// Allow other threads to run
	}
}


unsigned long WINAPI foo_bluetooth_ctrl::read_command_thread(LPVOID param)
{
	foo_bluetooth_ctrl *ctrl = reinterpret_cast<foo_bluetooth_ctrl*>(param);
	ctrl->read_command();
	return 0;
}


LRESULT CALLBACK foo_bluetooth_ctrl::wndproc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
	case WM_BT_CMD:
		// global instance.. maybe we should use singleton?
		g_bt_ctrl->m_bt_control->send_command(lParam);
		break;
	}
	return uDefWindowProc(hwnd,msg,wParam,lParam);
}








/****************
 * CONFIG dialog
 ****************/
BOOL CALLBACK config_bluetooth_ctrl::ConfigProc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp)
{
	string8 tmpstr,tmp2str;
	switch(msg)
	{
	case WM_INITDIALOG:
		uSetDlgItemText(wnd, IDC_MUSICDIR, cfg_musicdir);
		uSetDlgItemText(wnd, IDC_COMPORT, cfg_comport);
		uSetDlgItemText(wnd, IDC_PLAYLIST_FMT, cfg_playlist_fmt);
		uSendDlgItemMessage(wnd,IDC_AUTOCONNECT,BM_SETCHECK,cfg_autoconnect.get_value(),0);
		uSendDlgItemMessage(wnd,IDC_RANDOM,BM_SETCHECK,cfg_random.get_value(),0);
		EnableWindow(wnd, TRUE);
		break;

	case WM_DESTROY:
		break;

	case WM_COMMAND:
		switch(wp) {
		case IDC_COMPORT | (EN_CHANGE<<16):
			uGetDlgItemText(wnd, IDC_COMPORT, tmpstr);
			uStringUpper(tmp2str, tmpstr.get_ptr(), tmpstr.length());
			cfg_comport = tmp2str;
			break;

		case IDC_CONNECT_BUTTON:
			uGetDlgItemText(wnd, IDC_COMPORT, tmpstr);
			uStringUpper(tmp2str, tmpstr.get_ptr(), tmpstr.length());
			cfg_comport = tmp2str;
			// reconnect
			g_bt_ctrl->set_active(false);
			g_bt_ctrl->set_active(true);
			break;

		case IDC_MUSICDIR_BUTTON:
			uGetDlgItemText(wnd,IDC_MUSICDIR,tmpstr);
			if (uBrowseForFolder(wnd,"Select music directory",tmpstr))
			{
				tmp2str = cfg_musicdir.get_val();
				if(tmp2str.length() > 0)
					tmp2str.add_char(';');
				tmp2str.add_string(tmpstr);
				uSetDlgItemText(wnd,IDC_MUSICDIR,tmp2str);
				cfg_musicdir = tmp2str;
				g_bt_ctrl->update_config();
			}			
			break;

		case IDC_MUSICDIR | (EN_CHANGE<<16):
			uGetDlgItemText(wnd, IDC_MUSICDIR, tmpstr);
			cfg_musicdir = tmpstr;
			g_bt_ctrl->update_config();
			break;

		case IDC_AUTOCONNECT:
			cfg_autoconnect.set_value(uSendMessage((HWND)lp, BM_GETCHECK, 0, 0));
			break;

		case IDC_RANDOM:
			cfg_random.set_value(uSendMessage((HWND)lp, BM_GETCHECK, 0, 0));
			g_bt_ctrl->update_config();
			break;

		case IDC_PLAYLIST_FMT | (EN_CHANGE<<16):
            uGetDlgItemText(wnd, IDC_PLAYLIST_FMT, tmpstr);
			cfg_playlist_fmt = tmpstr;
			g_bt_ctrl->update_config();
			break;

		case IDC_TAGHELP:
			menu_manager::run_command("Foobar2000/Titleformatting help");
			break;
		}
		break;
	}
	return 0;
}
